import numpy as np
from gym.spaces import Box, Discrete


def get_tile_shape(shape, stack_size):
    obs_dim = len(shape)

    if obs_dim == 1:
        tile_shape = (stack_size,)
        new_shape = shape
    elif obs_dim == 3:
        tile_shape = (1,1,stack_size)
        new_shape = shape
    # stack 2-D frames
    elif obs_dim == 2:
        tile_shape = (1, 1, stack_size)
        new_shape = shape + (1,)
    else:
        assert False, "Stacking is only avaliable for 1,2 or 3 dimentional arrays"

    return tile_shape, new_shape

def stack_obs_space(obs_space, stack_size):
    '''
    obs_space_dict: Dictionary of observations spaces of agents
    stack_size: Number of frames in the observation stack
    Returns:
        New obs_space_dict
    '''
    if isinstance(obs_space, Box):
        dtype = obs_space.dtype
        obs_dim = obs_space.low.ndim
        # stack 1-D frames and 3-D frames
        tile_shape, new_shape = get_tile_shape(obs_space.low.shape, stack_size)

        low = np.tile(obs_space.low.reshape(new_shape), tile_shape)
        high = np.tile(obs_space.high.reshape(new_shape), tile_shape)
        new_obs_space = Box(low=low, high=high, dtype=dtype)
        return new_obs_space
    elif isinstance(obs_space, Discrete):
        return Discrete(obs_space.n ** stack_size)
    else:
         assert False, "Stacking is currently only allowed for Box and Discrete observation spaces. The given observation space is {}".format(obs_space)

def stack_init(obs_space, stack_size):
    if isinstance(obs_space, Box):
        tile_shape, new_shape = get_tile_shape(obs_space.low.shape, stack_size)
        return np.tile(np.zeros(new_shape),tile_shape)
    else:
        return 0

def stack_obs(frame_stack, obs, obs_space, stack_size):
    '''
    Parameters
    ----------
    frame_stack : if not None, it is the stack of frames
    obs : new observation
        Rearranges frame_stack. Appends the new observation at the end.
        Throws away the oldest observation.
    stack_size : needed for stacking reset observations
    '''
    if isinstance(obs_space, Box):
        obs_shape = obs.shape
        agent_fs = frame_stack

        if len(obs_shape) == 1:
            size = obs_shape[0]
            agent_fs[:-size] = agent_fs[size:]
            agent_fs[-size:] = obs
        elif len(obs_shape) == 2:
            agent_fs[:,:,:-1] = agent_fs[:,:,1:]
            agent_fs[:,:,-1] = obs
        elif len(obs_shape) == 3:
            nchannels = obs_shape[-1]
            agent_fs[:, :, :-nchannels] = agent_fs[:, :, nchannels:]
            agent_fs[:, :, -nchannels:] = obs
        return agent_fs
    elif isinstance(obs_space, Discrete):
        return (frame_stack * obs_space.n + obs) % (obs_space.n**stack_size)



import numpy as np
from gym.spaces import Box
from . import convert_box


def check_param(obs_space, shape):
    assert isinstance(shape, tuple), "shape must be tuple. It is {}".format(shape)
    assert all(isinstance(el,int) for el in shape), "shape must be tuple of ints, is: {}".format(shape)
    assert np.prod(shape) == np.prod(obs_space.shape), "new shape {} must have as many elements as original shape {}".format(shape,obs_space.shape)

def change_obs_space(obs_space, param):
    return convert_box(lambda obs:change_observation(obs, obs_space, param), obs_space)

def change_observation(obs, obs_space, shape):
    obs = obs.reshape(shape)
    return obs



from pettingzoo.tests import api_test,seed_test,error_tests
from pettingzoo.mpe import simple_push_v0
#from pettingzoo.sisl import multiwalker

from supersuit import continuous_actions,frame_stack
import numpy as np

def test_pettinzoo_frame_stack():
    _env = simple_push_v0.env()
    wrapped_env = frame_stack(_env)
    api_test.api_test(wrapped_env)
    # TODO: uncomment this when error tests is more stable
    # error_tests.error_test(frame_stack(simple_push_v0.env()))

def test_pettinzoo_continuous_actions():
    _env = simple_push_v0.env()
    wrapped_env = continuous_actions(_env)
    # wrapped_env.reset()
    # wrapped_env.step(np.nan*np.ones(4672))
    api_test.api_test(wrapped_env)
    seed_test.seed_test(lambda : continuous_actions(simple_push_v0.env()))



from supersuit.frame_stack import stack_obs_space,stack_init,stack_obs
from gym.spaces import Box, Discrete
import numpy as np
import pytest

stack_obs_space_3d = Box(low=np.float32(0.),high=np.float32(1.),shape=(4,4,3))
stack_obs_space_2d = Box(low=np.float32(0.),high=np.float32(1.),shape=(4,3))
stack_obs_space_1d = Box(low=np.float32(0.),high=np.float32(1.),shape=(3,))

stack_discrete = Discrete(3)

STACK_SIZE = 11

def test_obs_space():
    assert stack_obs_space(stack_obs_space_1d, STACK_SIZE).shape == (3*STACK_SIZE,)
    assert stack_obs_space(stack_obs_space_2d, STACK_SIZE).shape == (4,3,STACK_SIZE)
    assert stack_obs_space(stack_obs_space_3d, STACK_SIZE).shape ==  (4,4,3*STACK_SIZE)
    assert stack_obs_space(stack_discrete, STACK_SIZE).n == 3**STACK_SIZE

def stack_obs_helper(frame_stack_list, obs_space, stack_size):
    stack = stack_init(obs_space, stack_size)#stack_reset_obs(frame_stack_list[0], stack_size)
    for obs in frame_stack_list:
        stack = stack_obs(stack,obs,obs_space,stack_size)
    return stack

def test_change_observation():
    assert stack_obs_helper([stack_obs_space_1d.low],stack_obs_space_1d,STACK_SIZE).shape == (3*STACK_SIZE,)
    assert stack_obs_helper([stack_obs_space_1d.low,stack_obs_space_1d.high],stack_obs_space_1d,STACK_SIZE).shape == (3*STACK_SIZE,)
    assert stack_obs_helper([stack_obs_space_2d.low],stack_obs_space_2d,STACK_SIZE).shape == (4,3,STACK_SIZE)
    assert stack_obs_helper([stack_obs_space_2d.low,stack_obs_space_2d.high],stack_obs_space_2d,STACK_SIZE).shape == (4,3,STACK_SIZE)
    assert stack_obs_helper([stack_obs_space_3d.low],stack_obs_space_3d,STACK_SIZE).shape == (4,4,3*STACK_SIZE)

    assert stack_obs_helper([1,2],stack_discrete,STACK_SIZE) == 2 + 1 * 3

    stacked = stack_obs_helper([stack_obs_space_2d.low,stack_obs_space_2d.high],stack_obs_space_2d,3)
    raw = np.stack([np.zeros_like(stack_obs_space_2d.high),stack_obs_space_2d.low,stack_obs_space_2d.high],axis=2)
    assert np.all(np.equal(stacked, raw))

    stacked = stack_obs_helper([stack_obs_space_3d.low,stack_obs_space_3d.high],stack_obs_space_3d,3)
    raw = np.concatenate([np.zeros_like(stack_obs_space_3d.high),stack_obs_space_3d.low,stack_obs_space_3d.high],axis=2)
    assert np.all(np.equal(stacked, raw))



import numpy as np
from gym.spaces import Box
from . import convert_box


def check_param(obs_space, should_flatten):
    assert isinstance(should_flatten, bool), "should_flatten must be bool. It is {}".format(should_flatten)

def change_obs_space(obs_space, param):
    return convert_box(lambda obs:change_observation(obs, obs_space, param), obs_space)

def change_observation(obs, obs_space,should_flatten):
    if should_flatten:
        obs = obs.flatten()
    return obs



import numpy as np
import copy
from gym import spaces
import warnings
from skimage import measure
from .env import AECEnv

from .frame_stack import stack_obs_space, stack_obs

COLOR_RED_LIST = ["full", 'R', 'G', 'B']
OBS_RESHAPE_LIST = ["expand", "flatten"]


class wrapper(AECEnv):

    def __init__(self, env, color_reduction=None, down_scale=None, reshape=None, range_scale=None, new_dtype=None, continuous_actions=False, frame_stacking=1, homogenize_observations=False, homogenize_actions=False):
        '''
        Creates a wrapper around `env`.

        Parameters
        ----------
        env : TYPE
            DESCRIPTION.
        color_reduction : str or dict, optional
            True grayscale or reduce color to one channel. The default is None.
        down_scale : tuple or dict, optional
            Downscale observations using a filter. The default is None.
        reshape : str, optional
            Reshape observations: "expand" or "flatten". The default is None.
        range_scale : tuple or dict, optional
            new_obs = obs/(max_obs-min_obs) - min_obs. The default is None.
        new_dtype : type or dict, optional
            New dtype for observations. The default is None.
        continuous_actions : bool, optional
            If the action space is discrete, make a one-hot continuous vector. The default is False.
        frame_stacking : int, optional
            No. of observation frames to stack.. The default is 1.
        homogenize_observations: bool, optional
            If True, observations will be of same shape and will belong to the same observation_space
        homogenize_actions: bool, optional
            If True, actions will be of same shape and will belong to the same action_space

        Returns
        -------
        None.

        '''
        super().__init__()
        self.env = env
        self.color_reduction = color_reduction
        self.down_scale = down_scale
        self.reshape = reshape
        self.range_scale = range_scale
        self.new_dtype = new_dtype
        self.continuous_actions = continuous_actions
        self.frame_stacking = frame_stacking
        self.homogenize_observations = homogenize_observations
        self.homogenize_actions = homogenize_actions

        self.agents = self.env.agents
        self.agent_selection = self.env.agent_selection
        self.observation_spaces = self.env.observation_spaces
        self.action_spaces = copy.copy(self.env.action_spaces)
        self.orig_action_spaces = self.env.action_spaces

        self.rewards = self.env.rewards
        self.dones = self.env.dones
        self.infos = self.env.infos

        self.agent_order = self.env.agent_order

        if self.frame_stacking > 1:
            self.stack_of_frames = {agent: None for agent in self.agents}
            self.pre_fs_ndim = {agent: None for agent in self.agents}

        self._check_wrapper_params()

        self.modify_action_space()
        if self._check_box_space(self.observation_spaces):
            self.modify_observation_space()
        else:
            warnings.warn("All agents' observation spaces are not Box: {}, and as such the observation spaces are not modified.".format(self.observation_spaces))

        if self.homogenize_observations:
            self.original_observation_spaces = copy.deepcopy(self.observation_spaces)
            self.homogenize_observation_spaces()
        if self.homogenize_actions:
            self.original_action_spaces = copy.deepcopy(self.action_spaces)
            self.homogenize_action_spaces()

    def homogenize_observation_spaces(self):
        if self._check_box_space(self.observation_spaces):
            _flat_shapes = [(agent, np.product(self.observation_spaces[agent].shape)) for agent in self.observation_spaces]
            max_agent, _ = max(_flat_shapes, key=lambda x: x[1])  # gives you the first agent with the max length
            max_space = self.observation_spaces[max_agent]

            for agent in self.agents:
                if agent == max_agent:
                    continue
                _space = self.observation_spaces[agent]
                low = _space.low
                high = _space.high
                dtype = _space.dtype
                assert max_space.contains(low) and max_space.contains(high) and max_space.dtype == dtype, "Incompatible observation_spaces. Cannot homogenize."
                self.observation_spaces[agent] = copy.deepcopy(max_space)

        elif self._check_discrete_space(self.observation_spaces):
            _flat_shapes = [(agent, self.observation_spaces[agent].n) for agent in self.observation_spaces]
            max_agent, _ = max(_flat_shapes, key=lambda x: x[1])  # gives you the first agent with the max n
            max_n = self.observation_spaces[max_agent].n

            for agent in self.agents:
                if agent == max_agent:
                    continue
                if self.observation_spaces[agent].n is not max_n:
                    self.observation_spaces[agent] = spaces.Discrete(max_n)

        else:
            assert False, "Cannot homogenize observation_spaces. They are not gym.spaces.Box or gym.spaces.Discrete."

    def homogenize_observations_data(self, agent, obs):
        '''
            Pad observations with obs.low
        '''
        obs_space = self.observation_spaces[agent]

        if isinstance(obs_space, spaces.Box):
            new_obs = copy.copy(obs_space.low.flatten())
            new_obs[: np.product(obs.shape)] = obs
            return new_obs.reshape(obs_space.shape)

        elif isinstance(obs_space, spaces.Discrete):
            return obs
        else:
            return obs

    def dehomogenize_actions_data(self, agent, action):
        '''
            Extract the original actions
        '''
        orig_action_space = self.original_action_spaces[agent]

        if isinstance(orig_action_space, spaces.Box):
            # choose only the relevant action values
            new_action = action.flatten()
            new_action[: np.product(orig_action_space.shape)]
            return new_action.reshape(orig_action_space)

        elif isinstance(orig_action_space, spaces.Discrete):
            # extra action values refer to action value 0
            n = orig_action_space.n
            if action > n - 1:
                action = 0
            return action
        else:
            return action

    def homogenize_action_spaces(self):
        if self._check_box_space(self.action_spaces):
            _flat_shapes = [(agent, np.product(self.action_spaces[agent].shape)) for agent in self.action_spaces]
            max_agent, _ = max(_flat_shapes, key=lambda x: x[1])  # gives you the first agent with the max length
            max_space = self.action_spaces[max_agent]

            for agent in self.agents:
                if agent == max_agent:
                    continue
                _space = self.action_spaces[agent]
                low = _space.low
                high = _space.high
                dtype = _space.dtype
                assert max_space.contains(low) and max_space.contains(high) and max_space.dtype == dtype, "Incompatible action_spaces. Cannot homogenize."
                self.action_spaces[agent] = copy.deepcopy(max_space)

        elif self._check_discrete_space(self.action_spaces):
            _flat_shapes = [(agent, self.action_spaces[agent].n) for agent in self.action_spaces]
            max_agent, _ = max(_flat_shapes, key=lambda x: x[1])  # gives you the first agent with the max n
            max_n = self.action_spaces[max_agent].n

            for agent in self.agents:
                if agent == max_agent:
                    continue
                if self.action_spaces[agent].n is not max_n:
                    self.action_spaces[agent] = spaces.Discrete(max_n)

        else:
            assert False, "Cannot homogenize action_spaces. They are not gym.spaces.Box or gym.spaces.Discrete."

    def _check_wrapper_params(self):
        '''
        Checks if valid parameters are passed to wrapper object.

        Returns
        -------
        None.

        '''
        if self.color_reduction is not None:
            assert isinstance(self.color_reduction, str) or isinstance(self.color_reduction, dict), "color_reduction must be str or dict. It is {}".format(self.color_reduction)
            if isinstance(self.color_reduction, str):
                self.color_reduction = dict(zip(self.agents, [self.color_reduction for _ in enumerate(self.agents)]))
            if isinstance(self.color_reduction, dict):
                for agent in self.agents:
                    assert agent in self.color_reduction.keys(), "Agent id {} is not a key of color_reduction {}".format(agent, self.color_reduction)
                    assert self.color_reduction[agent] in COLOR_RED_LIST, "color_reduction must be in {}".format(COLOR_RED_LIST)
                    assert len(self.observation_spaces[agent].low.shape) == 3, "To apply color_reduction, length of shape of obs space of the agent should be 3. It is {}".format(len(self.observation_spaces[agent].low.shape))
                    if self.color_reduction[agent] == "full":
                        warnings.warn("You have chosen true grayscaling. It might be too slow. Choose a specific channel for better performance")

        if self.down_scale is not None:
            assert isinstance(self.down_scale, tuple) or isinstance(self.down_scale, dict), "down_scale must be tuple or dict. It is {}".format(self.down_scale)
            if isinstance(self.down_scale, tuple):
                self.down_scale = dict(zip(self.agents, [self.down_scale for _ in enumerate(self.agents)]))
            if isinstance(self.down_scale, dict):
                for agent in self.agents:
                    assert agent in self.down_scale.keys(), "Agent id {} is not a key of down_scale {}".format(agent, self.down_scale)

        if self.reshape is not None:
            assert self.reshape in OBS_RESHAPE_LIST, "reshape must be in {}".format(OBS_RESHAPE_LIST)

        if self.range_scale is not None:
            assert isinstance(self.range_scale, tuple) or isinstance(self.range_scale, dict) or isinstance(self.range_scale, str), "range_scale must be tuple or dict. It is {}".format(self.range_scale)
            if isinstance(self.range_scale, tuple):
                self.range_scale = dict(zip(self.agents, [self.range_scale for _ in enumerate(self.agents)]))
            if isinstance(self.range_scale, dict):
                for agent in self.agents:
                    assert agent in self.range_scale.keys(), "Agent id {} is not a key of range_scale {}".format(agent, self.range_scale)
                    assert len(self.range_scale[agent]) == 2, "Length of range_scale for agent {} is not 2.".format(agent)
                    assert self.range_scale[agent][0] <= self.range_scale[agent][1], "range_scale: for agent {}, low is greater than high".format(agent)

        assert isinstance(self.frame_stacking, int) and self.frame_stacking >= 1, "frame_stacking should be int and at least 1. It is {}".format(self.frame_stacking)
        assert isinstance(self.continuous_actions, bool), "continuous_actions is not a bool"

        if self.new_dtype is not None:
            assert isinstance(self.new_dtype, type) or isinstance(self.new_dtype, dict), "new_dtype must be type or dict. It is {}".format(self.new_dtype)
            if isinstance(self.new_dtype, type):
                self.new_dtype = dict(zip(self.agents, [self.new_dtype for _ in enumerate(self.agents)]))
            if isinstance(self.new_dtype, dict):
                for agent in self.agents:
                    assert agent in self.new_dtype.keys(), "Agent id {} is not a key of new_dtype {}".format(agent, self.new_dtype)
                    assert isinstance(self.new_dtype[agent], type), "new_dtype[agent] must be a dict of types. It is {}".format(self.new_dtype[agent])

    def _check_box_space(self, env_spaces):
        '''
        Checks if all (observation_ or action_) spaces are Box.

        Returns
        -------
        True if all spaces are gym.spaces.Box.
        '''
        return all([isinstance(space, spaces.Box) for space in env_spaces.values()])

    def _check_discrete_space(self, env_spaces):
        '''
        Checks if all (observation_ or action_) spaces are Discrete.

        Returns
        -------
        True if all spaces are gym.spaces.Discrete.
        '''
        return all([isinstance(space, spaces.Discrete) for space in env_spaces.values()])

    def modify_action_space(self):
        if self.continuous_actions:
            for agent in self.agents:
                act_space = self.orig_action_spaces[agent]
                if isinstance(act_space, spaces.Discrete):
                    new_act_space = spaces.Box(low=-np.inf, high=np.inf, shape=(act_space.n,))
                elif isinstance(act_space, spaces.MultiDiscrete):
                    new_act_space = spaces.Box(low=-np.inf, high=np.inf, shape=(np.sum(act_space.nvec),))
                elif isinstance(act_space, spaces.Box):
                    new_act_space = act_space
                else:
                    assert False, "space {} is not supported by the continuous_actions option of the wrapper".format(act_space)

                self.action_spaces[agent] = new_act_space
            print("Mod action space: continuous_actions", self.action_spaces)

    def modify_action(self, agent, action):
        new_action = action
        if self.continuous_actions:
            act_space = self.orig_action_spaces[agent]
            warped_act_space = self.action_spaces[agent]
            action = np.asarray(action)
            assert warped_act_space.contains(action), "received invalid action"

            def softmax(x):
                e_x = np.exp(x - np.max(x))
                return e_x / e_x.sum()

            def sample_softmax(vec):
                vec = vec.astype(np.float64)
                return np.argmax(np.random.multinomial(1, softmax(vec)))

            if isinstance(act_space, spaces.Discrete):
                new_action = sample_softmax(action)
            elif isinstance(act_space, spaces.MultiDiscrete):
                new_action = []
                sidx = 0
                for size in act_space.nvec:
                    samp = sample_softmax(action[sidx:(sidx + size)])
                    new_action.append(samp)
                    sidx = sidx + size
                new_action = np.asarray(new_action, dtype=np.int64)
            elif isinstance(act_space, spaces.Box):
                new_action = action
            else:
                assert False, "space {} is not supported by the continuous_actions option of the wrapper".format(act_space)

        return new_action

    def _color_reduction_obs_space(self):
        for agent in self.agents:
            obs_space = self.observation_spaces[agent]
            dtype = obs_space.dtype
            color_reduction = self.color_reduction[agent]
            if color_reduction == 'R':
                low = obs_space.low[:, :, 0]
                high = obs_space.high[:, :, 0]
            if color_reduction == 'G':
                low = obs_space.low[:, :, 1]
                high = obs_space.high[:, :, 1]
            if color_reduction == 'B':
                low = obs_space.low[:, :, 2]
                high = obs_space.high[:, :, 2]
            if color_reduction == 'full':
                low = np.average(obs_space.low, weights=[0.299, 0.587, 0.114], axis=2).astype(obs_space.dtype)
                high = np.average(obs_space.high, weights=[0.299, 0.587, 0.114], axis=2).astype(obs_space.dtype)
            self.observation_spaces[agent] = spaces.Box(low=low, high=high, dtype=dtype)
        print("Mod obs space: color_reduction", self.observation_spaces)

    def _down_scale_obs_space(self):
        for agent in self.agents:
            obs_space = self.observation_spaces[agent]
            dtype = obs_space.dtype
            down_scale = self.down_scale[agent]
            self.dtype_before_down_scale = copy.copy(dtype)
            mean = lambda x, axis: np.mean(x, axis=axis, dtype=dtype)
            low = measure.block_reduce(obs_space.low, block_size=down_scale, func=mean)
            high = measure.block_reduce(obs_space.high, block_size=down_scale, func=mean)
            self.observation_spaces[agent] = spaces.Box(low=low, high=high, dtype=dtype)
        print("Mod obs space: down_scale", self.observation_spaces)

    def _reshape_obs_space(self):
        for agent in self.agents:
            obs_space = self.observation_spaces[agent]
            reshape = self.reshape
            dtype = obs_space.dtype
            if reshape is OBS_RESHAPE_LIST[0]:
                # expand dim by 1
                low = np.expand_dims(obs_space.low, axis=-1)
                high = np.expand_dims(obs_space.high, axis=-1)
            elif reshape is OBS_RESHAPE_LIST[1]:
                # flatten
                low = obs_space.low.flatten()
                high = obs_space.high.flatten()
            self.observation_spaces[agent] = spaces.Box(low=low, high=high, dtype=dtype)
        print("Mod obs space: reshape", self.observation_spaces)

    def _range_scale_obs_space(self):
        for agent in self.agents:
            obs_space = self.observation_spaces[agent]
            if self.new_dtype is not None:
                dtype = self.new_dtype[agent]
            else:
                warnings.warn("Trying to scale observation_space, but a new dtype is not given. Defaulting to np.float32. Please verify if this is valid for your case.")
                dtype = np.float32
            shape = obs_space.shape
            self.observation_spaces[agent] = spaces.Box(low=0, high=1, shape=shape, dtype=dtype)
        print("Mod obs space: range_scale", self.observation_spaces)

    def _new_dtype_obs_space(self):
        for agent in self.agents:
            obs_space = self.observation_spaces[agent]
            dtype = self.new_dtype[agent]
            low = obs_space.low
            high = obs_space.high
            self.observation_spaces[agent] = spaces.Box(low=low, high=high, dtype=dtype)
        print("Mod obs space: new_dtype", self.observation_spaces)

    def _frame_stacking_obs_space(self):
        self.pre_fs_ndim = {agent: self.observation_spaces[agent].low.ndim for agent in self.agents}
        self.observation_spaces = stack_obs_space(self.observation_spaces, self.frame_stacking)
        print("Mod obs space: frame_stacking", self.observation_spaces)

    def modify_observation_space(self):
        # reduce color channels to 1
        if self.color_reduction is not None:
            self._color_reduction_obs_space()

        # downscale (image, typically)
        if self.down_scale is not None:
            self._down_scale_obs_space()

        # expand dimensions by 1 or flatten the array
        if self.reshape is not None:
            self._reshape_obs_space()

        # scale observation value (to [0,1], typically) and change observation_space dtype
        if self.range_scale is not None:
            self._range_scale_obs_space()
        elif self.new_dtype is not None:
            self._new_dtype_obs_space()

        if self.frame_stacking > 1:
            self._frame_stacking_obs_space()

    def _color_reduction_obs(self, obs, agent):
        color_reduction = self.color_reduction[agent]
        if color_reduction == 'R':
            obs = obs[:, :, 0]
        if color_reduction == 'G':
            obs = obs[:, :, 1]
        if color_reduction == 'B':
            obs = obs[:, :, 2]
        if color_reduction == 'full':
            obs = np.average(obs, weights=[0.299, 0.587, 0.114], axis=2).astype(obs.dtype)
        return obs

    def _down_scale_obs(self, obs, agent):
        down_scale = self.down_scale[agent]
        mean = lambda x, axis: np.mean(x, axis=axis, dtype=self.dtype_before_down_scale)
        obs = measure.block_reduce(obs, block_size=down_scale, func=mean)
        return obs

    def _reshape_obs(self, obs, agent):
        reshape = self.reshape
        if reshape is OBS_RESHAPE_LIST[0]:
            # expand dim by 1
            obs = np.expand_dims(obs, axis=-1)
        elif reshape is OBS_RESHAPE_LIST[1]:
            # flatten
            obs = obs.flatten()
        return obs

    def _range_scale_obs(self, obs, agent):
        range_scale = self.range_scale[agent]
        if self.new_dtype is not None:
            dtype = self.new_dtype[agent]
        else:
            warnings.warn("Trying to scale observation, but a new dtype is not given. Defaulting to np.float32. Please verify if this is valid for your case.")
            dtype = np.float32
        min_obs, max_obs = range_scale
        obs = np.divide(np.subtract(obs, min_obs), max_obs - min_obs, dtype=dtype)
        return obs

    def _new_dtype_obs(self, obs, agent):
        dtype = self.new_dtype[agent]
        obs = obs.astype(dtype)
        return obs

    def _frame_stacking_obs(self, obs, agent):
        stack_obs(self.stack_of_frames, agent, obs, self.frame_stacking)
        obs = self.stack_of_frames[agent]
        return obs

    def modify_observation(self, agent, observation):
        obs = observation
        # reduce color channels to 1
        if self.color_reduction is not None:
            obs = self._color_reduction_obs(obs, agent)

        # downscale (image, typically)
        if self.down_scale is not None:
            obs = self._down_scale_obs(obs, agent)

        # expand dimensions by 1 or flatten the array
        if self.reshape is not None:
            obs = self._reshape_obs(obs, agent)

        # scale observation value (to [0,1], typically) and change observation_space dtype
        if self.range_scale is not None:
            obs = self._range_scale_obs(obs, agent)
        elif self.new_dtype is not None:
            obs = self._new_dtype_obs(obs, agent)

        # frame_stacking
        if self.frame_stacking > 1:
            obs = self._frame_stacking_obs(obs, agent)
        return obs

    def close(self):
        self.env.close()

    def render(self, mode='human'):
        self.env.render(mode)

    def reset(self, observe=True):
        if observe:
            obs = self.env.reset(observe=True)
            agent = self.env.agent_selection
            observation = self.modify_observation(agent, obs)
            return observation
        else:
            self.env.reset(observe=False)

    def observe(self, agent):
        obs = self.env.observe(agent)
        observation = self.modify_observation(agent, obs)
        if self.homogenize_observations:
            observation = self.homogenize_observations_data(agent, observation)
        return observation

    def step(self, action, observe=True):
        agent = self.env.agent_selection
        if self.homogenize_actions:
            action = self.dehomogenize_actions_data(agent, action)
        action = self.modify_action(agent, action)
        if observe:
            next_obs = self.env.step(action, observe=True)

            observation = self.modify_observation(agent, next_obs)
        else:
            self.env.step(action, observe=False)
            observation = None

        self.agent_selection = self.env.agent_selection
        self.rewards = self.env.rewards
        self.dones = self.env.dones
        self.infos = self.env.infos

        return observation



from supersuit.basic_transforms.color_reduction import check_param,change_observation
from gym.spaces import Box
import numpy as np
import pytest

test_obs_space = Box(low=np.float32(0.),high=np.float32(1.),shape=(4,4,3))
bad_test_obs_space = Box(low=np.float32(0.),high=np.float32(1.),shape=(4,4,4))
test_obs = np.zeros([4,4,3])+np.arange(3)

def test_param_check():
    with pytest.raises(AssertionError):
        check_param(test_obs_space,"bob")
    with pytest.raises(AssertionError):
        check_param(bad_test_obs_space,"R")
    check_param(test_obs_space,"G")

def test_change_observation():
    new_obs = change_observation(test_obs,test_obs_space,"B")
    print(new_obs)
    assert np.all(np.equal(new_obs,2*np.ones([4,4])))



from supersuit.basic_transforms.down_scale import check_param,change_observation
from gym.spaces import Box
import numpy as np
import pytest

test_shape = (6,4,3)
high_val = (np.ones(test_shape) + np.arange(4).reshape(1,4,1)).astype(np.float32)
test_obs_space = Box(low=high_val-1,high=high_val)
test_val = high_val-0.5

def test_param_check():
    check_param(test_obs_space, (2,2,2))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, (2,2))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, (-2,2,2))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, (2.,2,2))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, 2)
    with pytest.raises(AssertionError):
        check_param(test_obs_space, ("B","C","D"))

def test_change_observation():
    new_obs = change_observation(test_val,test_obs_space,(3,2,2))

    test_obs = np.array([
        [[1.,  0.5],
        [3.,  1.5]],

        [[1.,  0.5],
        [3.,  1.5]]
    ])
    assert np.all(np.equal(new_obs,test_obs))



import warnings
import numpy as np
from gym.spaces import Box,Discrete
from gym import spaces

def check_action_space(act_space):
    assert isinstance(act_space, spaces.Discrete) or isinstance(act_space, spaces.Box),"space {} is not supported by the continuous_actions option of the wrapper".format(act_space)

def change_action_space(act_space):
    if isinstance(act_space, spaces.Discrete):
        new_act_space = spaces.Box(low=-np.float32(np.inf), high=np.float32(np.inf), shape=(act_space.n,))
    elif isinstance(act_space, spaces.Box):
        new_act_space = act_space

    return new_act_space

def modify_action(act_space, action, np_random):
    new_action = action

    def softmax(x):
        e_x = np.exp(x - np.max(x))
        return e_x / e_x.sum()

    def sample_softmax(vec):
        vec = vec.astype(np.float64)
        return np.argmax(np_random.multinomial(1, softmax(vec)))

    if isinstance(act_space, spaces.Discrete):
        if np.any(np.isnan(action)):
            new_action = np.nan
        else:
            new_action = sample_softmax(action)
    elif isinstance(act_space, spaces.Box):
        new_action = action

    return new_action



from .aec_wrappers import observation_lambda_wrapper,color_reduction,down_scale,dtype,flatten,reshape,normalize_obs,homogenize_obs,frame_stack,action_lambda_wrapper,homogenize_actions,continuous_actions



from gym.spaces import Box

def convert_box(convert_obs_fn, old_box):
    new_low = convert_obs_fn(old_box.low)
    new_high = convert_obs_fn(old_box.high)
    return Box(low=new_low, high=new_high, dtype=new_low.dtype)



from .base_aec_wrapper import BaseWrapper
from gym.spaces import Box,Space,Discrete
from . import basic_transforms
from .frame_stack import stack_obs_space,stack_init,stack_obs
from .action_transforms import homogenize_ops
from .action_transforms import continuous_action_ops
import numpy as np


class ObservationWrapper(BaseWrapper):
    def _modify_action(self, agent, action):
        return action

    def _update_step(self, agent, observation):
        pass

class observation_lambda_wrapper(ObservationWrapper):
    def __init__(self, env, change_observation_fn, change_obs_space_fn=None):
        assert callable(change_observation_fn), "change_observation_fn needs to be a function. It is {}".format(change_observation_fn)
        assert change_obs_space_fn is None or callable(change_obs_space_fn), "change_obs_space_fn needs to be a function. It is {}".format(change_obs_space_fn)
        self.change_observation_fn = change_observation_fn
        self.change_obs_space_fn = change_obs_space_fn

        super().__init__(env)

    def _check_wrapper_params(self):
        if self.change_obs_space_fn is None:
            spaces = self.observation_spaces.values()
            for space in spaces:
                assert isinstance(space, Box), "the observation_lambda_wrapper only allows the change_obs_space_fn argument to be optional for Box observation spaces"

    def _modify_spaces(self):
        new_spaces = {}
        for agent,space in self.observation_spaces.items():
            if self.change_obs_space_fn is None:
                new_low = self.change_observation_fn(space.low)
                new_high = self.change_observation_fn(space.high)
                new_spaces[agent] = Box(low=new_low, high=new_high, dtype=new_low.dtype)
            else:
                new_space = self.change_obs_space_fn(space)
                assert isinstance(new_space, Space), "output of change_obs_space_fn to observation_lambda_wrapper must be a gym space"
                new_spaces[agent] = new_space
        self.observation_spaces = new_spaces

    def _modify_observation(self, agent, observation):
        return self.change_observation_fn(observation)


class BasicObservationWrapper(ObservationWrapper):
    '''
    For internal use only
    '''
    def __init__(self,env,module,param):
        self.module = module
        self.param = param
        super().__init__(env)

    def _check_wrapper_params(self):
        assert all([isinstance(obs_space, Box) for obs_space in self.observation_spaces.values()]), \
            "All agents' observation spaces are not Box: {}, and as such the observation spaces are not modified.".format(self.observation_spaces)
        for obs_space in self.env.observation_spaces.values():
            self.module.check_param(obs_space,self.param)

    def _modify_spaces(self):
        new_spaces = {}
        for agent,space in self.observation_spaces.items():
            new_spaces[agent] = self.module.change_obs_space(space, self.param)
        self.observation_spaces = new_spaces

    def _modify_observation(self, agent, observation):
        obs_space = self.observation_spaces[agent]
        return self.module.change_observation(observation, obs_space, self.param)


class color_reduction(BasicObservationWrapper):
    def __init__(self,env,mode='full'):
        super().__init__(env,basic_transforms.color_reduction,mode)

class down_scale(BasicObservationWrapper):
    def __init__(self,env,x_scale=1,y_scale=1):
        old_obs_shape = list(env.observation_spaces.values())[0].shape
        scale_list = [1]*len(old_obs_shape)
        scale_list[0] = y_scale
        scale_list[0] = x_scale
        scale_tuple = tuple(scale_list)
        super().__init__(env,basic_transforms.down_scale,scale_tuple)

class dtype(BasicObservationWrapper):
    def __init__(self,env,dtype):
        super().__init__(env,basic_transforms.dtype,dtype)

class flatten(BasicObservationWrapper):
    def __init__(self,env):
        super().__init__(env,basic_transforms.flatten,True)

class reshape(BasicObservationWrapper):
    def __init__(self,env,shape):
        super().__init__(env,basic_transforms.reshape,shape)

class normalize_obs(BasicObservationWrapper):
    def __init__(self,env,env_min=0.,env_max=1.):
        shape = (env_min, env_max)
        super().__init__(env,basic_transforms.normalize_obs,shape)

class homogenize_obs(ObservationWrapper):
    def _check_wrapper_params(self):
        spaces = list(self.observation_spaces.values())
        homogenize_ops.check_homogenize_spaces(spaces)

    def _modify_spaces(self):
        spaces = list(self.observation_spaces.values())

        self._obs_space = homogenize_ops.homogenize_spaces(spaces)
        self.observation_spaces = {agent:self._obs_space for agent in self.observation_spaces}

    def _modify_observation(self, agent, observation):
        new_obs = homogenize_ops.homogenize_observations(self._obs_space,observation)
        return new_obs

class frame_stack(BaseWrapper):
    def __init__(self,env,num_frames=4):
        self.stack_size = num_frames
        super().__init__(env)

    def _check_wrapper_params(self):
        assert isinstance(self.stack_size, int), "stack size of frame_stack must be an int"
        for space in self.observation_spaces.values():
            if isinstance(space, Box):
                assert 1 <= len(space.shape) <= 3, "frame_stack only works for 1,2 or 3 dimentional observations"
            elif isinstance(space, Discrete):
                pass
            else:
                assert False, "Stacking is currently only allowed for Box and Discrete observation spaces. The given observation space is {}".format(obs_space)

    def reset(self, observe=True):
        self.stacks = {agent: stack_init(space, self.stack_size) for agent,space in self.env.observation_spaces.items()}
        return super().reset(observe)

    def _modify_spaces(self):
        self.observation_spaces = {agent: stack_obs_space(space, self.stack_size)  for agent,space in self.observation_spaces.items()}

    def _modify_action(self, agent, action):
        return action

    def _modify_observation(self, agent, observation):
        return self.stacks[agent]

    def _update_step(self, agent, observation):
        if observation is None:
            observation = self.observe(agent)
        self.stacks[agent] = stack_obs(self.stacks[agent], observation, self.env.observation_spaces[agent], self.stack_size)

class ActionWrapper(BaseWrapper):
    def __init__(self, env):
        super().__init__(env)

    def _modify_observation(self, agent, observation):
        return observation

    def _update_step(self, agent, observation):
        pass

class action_lambda_wrapper(ActionWrapper):
    def __init__(self, env, change_action_fn, change_space_fn):
        assert callable(change_action_fn), "change_action_fn needs to be a function. It is {}".format(change_action_fn)
        assert callable(change_space_fn), "change_space_fn needs to be a function. It is {}".format(change_space_fn)
        self.change_action_fn = change_action_fn
        self.change_space_fn = change_space_fn

        super().__init__(env)

    def _check_wrapper_params(self):
        pass

    def _modify_spaces(self):
        new_spaces = {}
        for agent,space in self.action_spaces.items():
            new_spaces[agent] = self.change_space_fn(space)
            assert isinstance(new_spaces[agent], Space), "output of change_space_fn argument to action_lambda_wrapper must be a gym space"

        self.action_spaces = new_spaces

    def _modify_action(self, agent, action):
        return self.change_action_fn(action, self.action_spaces[agent])

class homogenize_actions(ActionWrapper):
    def _check_wrapper_params(self):
        homogenize_ops.check_homogenize_spaces(list(self.env.action_spaces.values()))

    def _modify_spaces(self):
        space = homogenize_ops.homogenize_spaces(list(self.env.action_spaces.values()))

        self.action_spaces = {agent:space for agent in self.action_spaces}

    def _modify_action(self, agent, action):
        new_action = homogenize_ops.dehomogenize_actions(self.env.action_spaces[agent], action)
        return new_action

class continuous_actions(ActionWrapper):
    def __init__(self, env):
        SEED = 0x601326ad
        self.np_random = np.random.RandomState(SEED)
        super().__init__(env)

    def _check_wrapper_params(self):
        for space in self.action_spaces.values():
            continuous_action_ops.check_action_space(space)

    def _modify_spaces(self):
        spaces = {agent: continuous_action_ops.change_action_space(act_space) for agent,act_space in self.env.action_spaces.items()}

        self.action_spaces = spaces

    def _modify_action(self, agent, action):
        act_space = self.env.action_spaces[agent]
        new_action = continuous_action_ops.modify_action(act_space, action, self.np_random)
        return new_action

    def _remove_infos(self):
        self.infos = {agent: {key:value for key,value in info_dict.items() if key != "legal_moves"}
            for agent,info_dict in self.infos.items()}

    def reset(self, observe=True):
        res = super().reset(observe)
        self._remove_infos()
        return res

    def step(self, action, observe=True):
        res = super().step(action, observe)
        self._remove_infos()
        return res



from skimage import measure
from gym.spaces import Box
import numpy as np
from . import convert_box


def check_param(obs_space, down_scale):
    assert isinstance(down_scale, tuple), "down_scale must be tuple. It is {}".format(down_scale)
    assert all(isinstance(ds, int) for ds in down_scale), "down_scale must be a tuple of int. It is {}".format(down_scale)
    assert len(obs_space.shape) == len(down_scale), "down scale must be of the same length as the observations of the agents"
    assert all(ds > 0 for ds in down_scale), "all down scale values must be greater than zero"

def change_obs_space(obs_space, param):
    return convert_box(lambda obs:change_observation(obs, obs_space, param), obs_space)

def change_observation(obs, obs_space,down_scale):
    mean = lambda x, axis: np.mean(x, axis=axis, dtype=obs_space.dtype)
    obs = measure.block_reduce(obs, block_size=down_scale, func=mean)
    return obs



from supersuit.basic_transforms.dtype import check_param,change_observation
from gym.spaces import Box
import numpy as np
import pytest

test_obs_space = Box(low=np.float32(0.),high=np.float32(1.),shape=(4,4,3),dtype=np.float32)
test_obs = np.zeros([4,4,3],dtype=np.float64)+np.arange(3)


def test_param_check():
    check_param(test_obs_space, np.uint8)
    check_param(test_obs_space, np.dtype("uint8"))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, 6)

def test_change_observation():
    new_obs = change_observation(test_obs,test_obs_space,np.float32)
    assert new_obs.dtype == np.float32



from supersuit.basic_transforms.normalize_obs import check_param,change_obs_space,change_observation
from gym.spaces import Box
import numpy as np
import pytest

high_val = np.array([1,2,4])
test_val = np.array([1,1,1])
test_obs_space = Box(low=np.zeros(3,dtype=np.float32),high=high_val.astype(np.float32))
bad_test_obs_space = Box(low=np.zeros(3,dtype=np.int32),high=high_val.astype(np.int32),dtype=np.int32)
bad_test_obs_space2 = Box(low=np.zeros(3,dtype=np.float32),high=np.inf*high_val.astype(np.float32))


def test_param_check():
    check_param(test_obs_space, (2,3))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, (2,2))
    with pytest.raises(AssertionError):
        check_param(test_obs_space, ("bob",2))
    with pytest.raises(AssertionError):
        check_param(bad_test_obs_space, (2,3))
    with pytest.raises(AssertionError):
        check_param(bad_test_obs_space2, (2,3))

def test_change_obs_space():
    assert np.all(np.equal(change_obs_space(test_obs_space,(1,2)).high, np.array([2,2,2])))

def test_change_observation():
    print(change_observation(test_val,test_obs_space,(1,2)))
    assert np.all(np.equal(change_observation(test_val,test_obs_space,(1,2)), np.array([2,1.5,1.25])))



import numpy as np
import copy
from gym.spaces import Box
from gym import spaces
import warnings
from skimage import measure
from pettingzoo import AECEnv


class BaseWrapper(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self, env):
        '''
        Creates a wrapper around `env`. Extend this class to create changes to the space.
        '''
        super().__init__()
        self.env = env

        self.num_agents = self.env.num_agents
        self.agents = self.env.agents
        self.observation_spaces = copy.copy(self.env.observation_spaces)
        self.action_spaces = copy.copy(self.env.action_spaces)
        self.agent_order = self.env.agent_order

        self._check_wrapper_params()

        self._modify_spaces()

    def _check_wrapper_params(self):
        raise NotImplementedError()

    def _modify_spaces(self):
        raise NotImplementedError()

    def _modify_action(self, agent, action):
        raise NotImplementedError()

    def _modify_observation(self, agent, observation):
        raise NotImplementedError()

    def _update_step(self, agent, observation):
        raise NotImplementedError()

    def close(self):
        self.env.close()

    def render(self, mode='human'):
        self.env.render(mode)

    def reset(self, observe=True):
        observation = self.env.reset(observe)
        agent = self.env.agent_selection

        self.agent_selection = self.env.agent_selection
        self.agent_order = self.env.agent_order
        self.rewards = self.env.rewards
        self.dones = self.env.dones
        self.infos = self.env.infos

        self._update_step(agent,observation)
        if observe:
            observation = self._modify_observation(agent,observation)
            return observation
        else:
            return None

    def observe(self, agent):
        obs = self.env.observe(agent)
        observation = self._modify_observation(agent, obs)
        return observation

    def step(self, action, observe=True):
        agent = self.env.agent_selection
        cur_act_space = self.action_spaces[agent]
        assert not isinstance(cur_act_space,Box) or cur_act_space.shape == action.shape, "the shape of the action {} is not equal to the shape of the action space {}".format(action.shape,cur_act_space.shape)
        action = self._modify_action(agent, action)
        next_obs = self.env.step(action, observe=observe)
        new_agent = self.env.agent_selection
        self._update_step(new_agent,next_obs)

        self.agent_selection = self.env.agent_selection
        self.agent_order = self.env.agent_order
        self.rewards = self.env.rewards
        self.dones = self.env.dones
        self.infos = self.env.infos

        if observe:
            next_obs = self._modify_observation(new_agent,next_obs)
            return next_obs
        else:
            return None



from gym.spaces import Box, Discrete
from .dummy_aec_env import DummyEnv
import numpy as np
from supersuit.aec_wrappers import frame_stack,reshape,observation_lambda_wrapper,action_lambda_wrapper,homogenize_actions,continuous_actions,homogenize_obs
from supersuit import aec_wrappers

base_obs = {"a{}".format(idx): np.zeros([8,8,3]) + np.arange(3) + idx for idx in range(2)}
base_obs_space = {"a{}".format(idx): Box(low=np.float32(0.),high=np.float32(10.),shape=[8,8,3]) for idx in range(2)}
base_act_spaces = {"a{}".format(idx): Discrete(5) for idx in range(2)}

def test_frame_stack():
    base_obs_space = {"a{}".format(idx): Box(low=np.float32(0.),high=np.float32(10.),shape=[2,3]) for idx in range(2)}
    base_obs = {"a{}".format(idx): np.zeros([2,3]) + np.arange(3) + idx for idx in range(2)}
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = frame_stack(base_env, 4)
    obs = env.reset()
    assert obs.shape == (2,3,4)
    first_obs = env.step(2)
    assert np.all(np.equal(first_obs[:,:,-1],base_obs["a1"]))
    assert np.all(np.equal(first_obs[:,:,:-1],0))

    base_obs = {"a{}".format(idx): idx+3 for idx in range(2)}
    base_env = DummyEnv(base_obs, base_act_spaces, base_act_spaces)
    env = frame_stack(base_env, 4)
    obs = env.reset()
    assert env.observation_spaces[env.agent_selection].n == 5**4
    first_obs = env.step(2)
    assert first_obs == 4
    second_obs = env.step(2)
    assert second_obs == 3+3*5
    for x in range(100):
        nth_obs = env.step(2)
    assert nth_obs == ((3*5+3)*5+3)*5+3



def test_reshape():
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = reshape(base_env, (64, 3))
    obs = env.reset()
    assert obs.shape == (64,3)
    first_obs = env.step(5)
    assert np.all(np.equal(first_obs,base_obs["a1"].reshape([64,3])))

def new_dummy():
    return  DummyEnv(base_obs, base_obs_space, base_act_spaces)

def test_basic_wrappers():
    wrappers = [
        aec_wrappers.color_reduction(new_dummy(),"R"),
        aec_wrappers.down_scale(new_dummy(),x_scale=5,y_scale=10),
        aec_wrappers.dtype(new_dummy(),np.int32),
        aec_wrappers.flatten(new_dummy()),
        aec_wrappers.reshape(new_dummy(),(64,3)),
        aec_wrappers.normalize_obs(new_dummy(),env_min=-1,env_max=5.),
        aec_wrappers.frame_stack(new_dummy(),8),
        aec_wrappers.homogenize_obs(new_dummy()),
    ]
    for env in wrappers:
        obs = env.reset()
        first_obs = env.observe("a1")


def test_lambda():
    def add1(obs):
        return obs+1
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = observation_lambda_wrapper(base_env, add1)
    obs0 = env.reset()
    assert int(obs0[0][0][0]) == 1
    env = observation_lambda_wrapper(env, add1)
    obs0 = env.reset()
    assert int(obs0[0][0][0]) == 2

    def tile_obs(obs):
        shape_size = len(obs.shape)
        tile_shape = [1]*shape_size
        tile_shape[0] *= 2
        return np.tile(obs,tile_shape)

    env = observation_lambda_wrapper(env, tile_obs)
    obs0 = env.reset()
    assert env.observation_spaces[env.agent_selection].shape == (16,8,3)
    def change_shape_fn(obs_space):
        return Box(low=0,high=1,shape=(32,8,3))
    env = observation_lambda_wrapper(env, tile_obs)
    obs0 = env.reset()
    assert env.observation_spaces[env.agent_selection].shape == (32,8,3)
    assert obs0.shape == (32,8,3)

def test_action_lambda():
    def inc1(x,space):
        return (x + 1)
    def change_space_fn(space):
        return Discrete(space.n+1)
    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = action_lambda_wrapper(base_env, inc1, change_space_fn)
    env.reset()
    env.step(5)

def test_dehomogenize():
    base_act_spaces = {"a{}".format(idx): Discrete(5+idx) for idx in range(2)}

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = homogenize_actions(base_env)
    env.reset()
    assert all([s.n == 6 for s in env.action_spaces.values()])
    env.step(5)

def test_continuous_actions():
    base_act_spaces = {"a{}".format(idx): Discrete(5) for idx in range(2)}

    base_env = DummyEnv(base_obs, base_obs_space, base_act_spaces)
    env = continuous_actions(base_env)
    env.reset()
    assert all([s.shape == (5,) for s in env.action_spaces.values()])
    env.step(np.ones(5))
    res = env.step(np.nan*np.ones(5))



from pettingzoo import AECEnv
import copy
from pettingzoo.utils.agent_selector import agent_selector

class DummyEnv(AECEnv):
    def __init__(self,observations,observation_spaces,action_spaces):
        super().__init__()
        self._observations = observations
        self.observation_spaces = observation_spaces

        self.agents = [x for x in observation_spaces.keys()]
        self.num_agents = len(self.agents)
        self.agent_order = list(self.agents)
        self._agent_selector = agent_selector(self.agent_order)
        self.agent_selection = self._agent_selector.reset()
        self.action_spaces = action_spaces

        self.rewards = {a:0 for a in self.agents}
        self.dones = {a:False for a in self.agents}
        self.infos = {a:{} for a in self.agents}

        self.agent_order = self.agents

    def observe(self, agent):
        return self._observations[agent]

    def step(self, action, observe=True):
        old_sel = self.agent_selection
        self.agent_selection = self._agent_selector.next()
        if observe:
            return self._observations[self.agent_selection]
        else:
            return None

    def reset(self, observe=True):
        self.agent_selection = self._agent_selector.reset()
        return self._observations[self.agent_selection]

    def close(self):
        pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SuperSuit",
    version="0.1.10",
    author="PettingZoo Team",
    author_email="justinkterry@gmail.com",
    description="Wrappers for Gym and PettingZoo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PettingZoo-Team/SuperSuit",
    keywords=["Reinforcement Learning", "gym"],
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    data_files = [("", ["LICENSE.txt"])],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)



from skimage import measure
from gym.spaces import Box
import numpy as np

def check_param(obs_space, min_max_pair):
    assert np.dtype(obs_space.dtype) == np.dtype("float32") or np.dtype(obs_space.dtype) == np.dtype("float64")
    assert isinstance(min_max_pair, tuple) and len(min_max_pair) == 2, "range_scale must be tuple of size 2. It is {}".format(min_max_pair)
    try:
        min_res = float(min_max_pair[0])
        max_res = float(min_max_pair[1])
    except ValueError:
        assert False, "normalize_obs inputs must be numbers. They are {}".format(min_max_pair)
    assert max_res > min_res, "maximum must be greater than minimum value in normalize_obs"
    assert np.all(np.isfinite(obs_space.low)) and np.all(np.isfinite(obs_space.high))

def change_obs_space(obs_space, min_max_pair):
    min = np.float64(min_max_pair[0]).astype(obs_space.dtype)
    max = np.float64(min_max_pair[1]).astype(obs_space.dtype)
    return Box(low=min,high=max,shape=obs_space.shape,dtype=obs_space.dtype)

def change_observation(obs, obs_space,  min_max_pair):
    min_res,max_res = [float(x) for x in min_max_pair]
    old_size = obs_space.high - obs_space.low
    new_size = max_res - min_res
    result = (obs - obs_space.low) / old_size * new_size + min_res
    return result



import warnings
import numpy as np
from gym.spaces import Box
from gym import spaces
from . import convert_box

COLOR_RED_LIST = ["full", 'R', 'G', 'B']
GRAYSCALE_WEIGHTS = [0.299, 0.587, 0.114]

def check_param(space,color_reduction):
    assert isinstance(color_reduction, str), "color_reduction must be str. It is {}".format(color_reduction)
    assert color_reduction in COLOR_RED_LIST, "color_reduction must be in {}".format(COLOR_RED_LIST)
    assert len(space.low.shape) == 3 and space.low.shape[2] == 3, "To apply color_reduction, shape must be a 3d image with last dimention of size 3. Shape is {}".format((space.low.shape))
    if color_reduction == "full":
        warnings.warn("You have chosen true grayscaling. It might be too slow. Choose a specific channel for better performance")

def change_obs_space(obs_space, param):
    return convert_box(lambda obs:change_observation(obs, obs_space, param), obs_space)

def change_observation(obs,  obs_space,color_reduction):
    if color_reduction == 'R':
        obs = obs[:, :, 0]
    if color_reduction == 'G':
        obs = obs[:, :, 1]
    if color_reduction == 'B':
        obs = obs[:, :, 2]
    if color_reduction == 'full':
        obs = np.average(obs, weights=GRAYSCALE_WEIGHTS, axis=2).astype(obs.dtype)
    return obs



import warnings
import numpy as np
from gym.spaces import Box,Discrete
from gym import spaces

def check_homogenize_spaces(all_spaces):
    assert len(all_spaces) > 0
    space1 = all_spaces[0]
    assert all(isinstance(space,space1.__class__) for space in all_spaces), "all spaces to homogenize must be of same general shape"

    if isinstance(space1, spaces.Box):
        for space in all_spaces:
            assert isinstance(space, spaces.Box), "all spaces for homogenize must be either Box or Discrete, not a mix"
            assert len(space1.shape) == len(space.shape), "all spaces to homogenize must be of same shape"
            assert space1.dtype == space.dtype, "all spaces to homogenize must be of same dtype"
    elif isinstance(space1, spaces.Discrete):
        for space in all_spaces:
            assert isinstance(space, spaces.Discrete), "all spaces for homogenize must be either Box or Discrete, not a mix"
    else:
        assert False, "homogenization only supports Discrete and Box spaces"

def pad_to(arr,new_shape,pad_value):
    old_shape = arr.shape
    pad_size = [ns-os for ns,os in zip(new_shape,old_shape)]
    pad_tuples = [(0,ps) for ps in pad_size]
    return np.pad(arr,pad_tuples,constant_values=pad_value)

def homogenize_spaces(all_spaces):
    space1 = all_spaces[0]
    print(space1.dtype)
    if isinstance(space1, spaces.Box):
        all_dims = np.array([space.shape for space in all_spaces],dtype=np.int32)
        max_dims = np.max(all_dims,axis=0)
        new_shape = tuple(max_dims)
        all_lows = np.stack([pad_to(space.low,new_shape,np.minimum(0,np.min(space.low))) for space in all_spaces])
        all_highs = np.stack([pad_to(space.high,new_shape,np.maximum(1e-5,np.max(space.high))) for space in all_spaces])
        new_low = np.min(all_lows,axis=0)
        new_high = np.max(all_highs,axis=0)
        assert new_shape == new_low.shape
        return Box(low=new_low,high=new_high,dtype=space1.dtype)
    elif isinstance(space1, spaces.Discrete):
        max_n = max([space.n for space in all_spaces])
        return Discrete(max_n)
    else:
        assert False

def dehomogenize_actions(orig_action_space, action):
    if isinstance(orig_action_space, spaces.Box):
        # choose only the relevant action values
        cur_shape = action.shape
        new_shape = orig_action_space.shape
        assert len(cur_shape) == len(new_shape)
        slices = [slice(0,i) for i in new_shape]
        new_action = action[tuple(slices)]

        return new_action

    elif isinstance(orig_action_space, spaces.Discrete):
        # extra action values refer to action value 0
        n = orig_action_space.n
        if action > n - 1:
            action = 0
        return action
    else:
        assert False

def homogenize_observations(obs_space, obs):
    if isinstance(obs_space, spaces.Box):
        return pad_to(obs,obs_space.shape,0)
    elif isinstance(obs_space, spaces.Discrete):
        return obs_space
    else:
        assert False
